@@ -1,12 +1,14 @@
-/* BeroList 1.0.2                                                 */
+/* BeroList 2.0.0                                                 */
 /* An easy-to-use mailing list server                             */
 /* (c) 1996 by Bernhard Rosenkraenzer <root@startrek.in-trier.de> */
 /* Fixed by Jarno Paananen <jpaana@kalahari.ton.tut.fi>           */
+/*          Stefano Torricella <stetor@ronchiato.it>              */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "list.h"
+#include "socket.h"
 #ifndef HOST
 	#include <unistd.h>
   	char *hostname;
@@ -16,6 +18,7 @@
 #ifdef LOG
 	FILE *log;
 #endif
+#define VALIDCHAR "\t\n\r\b\000"
 char *listaddress;
 
 /* Function prototypes... */
@@ -30,24 +33,30 @@ void unsubscribe(char *listname, char *user, char *operator);
 char *email(char *full_address);
 char *realname(char *full_address);
 char *username(char *address);
+char *downcase(char *s);
 int scasecmp(char *s1, char *s2); /* Fix a buggy implementation of strcasecmp */
 
 /* Functions... */
 
 char *extract(char *from, const char *begin, const char end)
 {
-  char *found;
+  char *found,*result;
 
-  if((char *)strstr(from,begin)!=NULL) {
-  	found=(char *) malloc(strlen(from) + 1);
-  	strcpy(found,(char *) strstr(from,begin)+strlen(begin));
-  	if((char *) strchr(found,end)!=NULL)
-  		found[strchr(found,end)-found]='\0';
-  	if((char *) strchr(found,'\n')!=NULL)
-  		found[strchr(found,'\n')-found]='\0';
-  	return(found);
-  } else
-  	return(NULL);
+  found=strstr(from,begin);
+
+  while(found){
+    if(strchr(VALIDCHAR,(int) (*(found-1))) || !(*(found-1))) {
+          result=(char *) malloc(strlen(from) + 1);
+          strcpy(result,(char *) found+strlen(begin));
+          if((char *) strchr(result,end)!=NULL)
+                  result[strchr(result,end)-result]='\0';
+          if((char *) strchr(result,'\n')!=NULL)
+                  result[strchr(result,'\n')-result]='\0';
+          return(result);
+    };
+    found=strstr(found+1,begin);
+  };
+  return(NULL);
 }
 
 char *readfile(char *filename)
@@ -84,7 +93,7 @@ char *readline(FILE *file)
   
   line=(char *) malloc(sizeof(char)*1024);
   last=0; len=0;
-  while(!feof(file) && last!='\n') {
+  while(!feof(file) && last!=EOF && last!='\n') {
   	last=(char) fgetc(file);
   	if(last!='\n' && !feof(file)) {
   		line[len]=last;
@@ -97,42 +106,98 @@ char *readline(FILE *file)
 
 void send(char *list, char *to, char *from, char *subject, char *message, char *footfile)
 {
-  FILE *mailer;
-  char *mailer_cmd,*footer;
+  char *footer;
+  #ifdef USE_SENDMAIL
+	FILE *mailer;
+	char *mailer_cmd;
+  #else
+  	int socket;
+  	char *server_reply;
+  #endif
   
   printf("Sending...\n");
-
-  #ifdef TRUSTED
-  	from=email(from);
-  	mailer_cmd=(char *) malloc(strlen(SENDMAIL)+strlen(from)+strlen(to)+7);
-	sprintf(mailer_cmd,"%s -f%s '%s'",SENDMAIL,from,to);
-	#ifdef LOG
-  		fprintf(log,"* %s\n",mailer_cmd);
-  	#endif
-  	mailer=popen(mailer_cmd,"w");
-  #else
-	mailer_cmd=(char *) malloc(strlen(SENDMAIL)+strlen(to)+4);
-	sprintf(mailer_cmd,"%s '%s'",SENDMAIL,to); 
-	#ifdef LOG
-		fprintf(log,"* %s\n",mailer_cmd);
+  #ifdef USE_SENDMAIL     /* USE_SENDMAIL: Use the 1.x.x way */
+	#ifdef TRUSTED
+		from=email(from);
+		mailer_cmd=(char *) malloc(strlen(SENDMAIL)+strlen(from)+strlen(to)+7);
+		sprintf(mailer_cmd,"%s -f%s '%s'",SENDMAIL,from,to);
+		#ifdef LOG
+	  		fprintf(log,"* %s\n",mailer_cmd);
+	  	#endif
+	  	mailer=popen(mailer_cmd,"w");
+	#else
+		mailer_cmd=(char *) malloc(strlen(SENDMAIL)+strlen(to)+4);
+		sprintf(mailer_cmd,"%s '%s'",SENDMAIL,to); 
+		#ifdef LOG
+			fprintf(log,"* %s\n",mailer_cmd);
+		#endif
+		mailer=popen(mailer_cmd,"w");
+		fprintf(mailer,"From: %s\n",from);
 	#endif
-	mailer=popen(mailer_cmd,"w");
-	fprintf(mailer,"From: %s\n",from);
-  #endif
-  if(subject==NULL && strncasecmp(message,"subject:",8)==0) {
-	/* Subject: line @ beginning of msg body... */
-	subject=(char *) malloc(strlen(message)-7);
-	strcpy(subject,message+8);
-	while(strchr(subject,' ')==subject) subject++;
-	subject[strchr(subject,' ')-subject]='\0';
-	message=strchr(message,'\n')+1;
-  }
-  if(subject!=NULL)	fprintf(mailer,"Subject: %s\n",subject);
-  fprintf(mailer,"\n%s\n",message);
-  footer=readfile(footfile);
-  if(footer!=NULL)	fprintf(mailer,"%s\n",footer);
-  fprintf(mailer,".\n");
-  pclose(mailer);  
+	if(subject==NULL && strncasecmp(message,"subject:",8)==0) {
+		/* Subject: line @ beginning of msg body... */
+		subject=(char *) malloc(strlen(message)-7);
+		strcpy(subject,message+8);
+		while(strchr(subject,' ')==subject) subject++;
+		subject[strchr(subject,' ')-subject]='\0';
+		message=strchr(message,'\n')+1;
+	}
+	if(subject!=NULL)	fprintf(mailer,"Subject: %s\n",subject);
+	fprintf(mailer,"\n%s\n",message);
+	footer=readfile(footfile);
+	if(footer!=NULL)	fprintf(mailer,"%s\n",footer);
+	fprintf(mailer,".\n");
+	pclose(mailer);
+  #else
+  	server_reply=(char *) malloc(4096);  /* I know, 4k might be exaggerated... */
+  	socket=Socket(SMTP_SERVER,25);
+  	if(socket<=0) {
+  		printf("Can't connect to server %s.\n",SMTP_SERVER);
+  		exit(1);
+  	}
+  	printf("Connected to %s.\n",SMTP_SERVER);
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+  	SockPrintf(socket,"HELO %s\r\n",HOST);
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+  	SockPrintf(socket,"MAIL FROM: %s\r\n",email(from));
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+	SockPrintf(socket,"RCPT TO: %s\r\n",email(to));
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	SockPrintf(socket,"DATA\r\n");
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	SockPrintf(socket,"From: %s\r\n",from);
+	SockPrintf(socket,"To: %s\r\n",to);
+	if(subject==NULL && strncasecmp(message,"subject:",8)==0) {
+		/* Subject: line @ beginning of msg body... */
+		subject=(char *) malloc(strlen(message)-7);
+		strcpy(subject,message+8);
+		while(strchr(subject,' ')==subject) subject++;
+		subject[strchr(subject,' ')-subject]='\0';
+		message=strchr(message,'\n')+1;
+	}
+	if(subject!=NULL)	SockPrintf(socket,"Subject: %s\r\n",subject);
+	SockPrintf(socket,"\n%s\n",message);
+	footer=readfile(footfile);
+	if(footer!=NULL)	SockPrintf(socket,"%s\n",footer);
+	SockPrintf(socket,".\n");
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	if(strstr(server_reply,"accepted")==NULL && strstr(server_reply,"Accepted")==NULL && strstr(server_reply,"ACCEPTED")==NULL) {
+		puts("Mail rejected by SMTP server.\n");
+		SockKill(socket);
+		exit(1);
+	}
+	SockPrintf(socket,"QUIT\r\n");
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	SockKill(socket);
+	free(server_reply);
+  #endif  
 }
 
 void sendtolist(char *listname, char *from, char *subject, char *message)
@@ -141,6 +206,12 @@ void sendtolist(char *listname, char *from, char *subject, char *message)
   char *oldmsg;
   char *config,*sender;
   FILE *list;
+  #ifndef USE_SENDMAIL
+  	int socket;
+  	char *server_reply;
+  	char *footer;
+  	char success;
+  #endif
 
   printf("Sending to list...\n");
   
@@ -174,12 +245,67 @@ void sendtolist(char *listname, char *from, char *subject, char *message)
   #endif
 
   list=fopen(listfile,"r");
-  while(!feof(list)) {
-  	member=readline(list);
-  	if(strlen(member)>2)
-  		send(listname,member,sender,subject,message,footfile);
-  	free(member);
-  }
+
+  #ifdef USE_SENDMAIL
+	while(!feof(list)) {
+		member=readline(list);
+		if(strlen(member)>2)
+			send(listname,member,sender,subject,message,footfile);
+		free(member);
+	}
+  #else
+  	server_reply=(char *) malloc(4096);  /* I know, 4k might be exaggerated... */
+  	socket=Socket(SMTP_SERVER,25);
+  	if(socket<=0) {
+  		printf("Can't connect to server %s.\n",SMTP_SERVER);
+  		exit(1);
+  	}
+  	printf("Connected to %s.\n",SMTP_SERVER);
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+  	SockPrintf(socket,"HELO %s\r\n",HOST);
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+  	SockPrintf(socket,"MAIL FROM: %s\r\n",email(sender));
+  	SockGets(socket,server_reply,4096);
+  	puts(server_reply);
+  	while(!feof(list)) {
+  		member=readline(list);
+  		if(strlen(member)>2) {
+  			SockPrintf(socket,"RCPT TO: %s\r\n",email(member));
+  			SockGets(socket,server_reply,4096);
+  			puts(server_reply);
+  		}
+  	}
+	SockPrintf(socket,"DATA\r\n");
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	SockPrintf(socket,"From: %s\r\n",sender);
+	SockPrintf(socket,"To: %s\r\n",listaddress);
+	if(subject==NULL && strncasecmp(message,"subject:",8)==0) {
+		/* Subject: line @ beginning of msg body... */
+		subject=(char *) malloc(strlen(message)-7);
+		strcpy(subject,message+8);
+		while(strchr(subject,' ')==subject) subject++;
+		subject[strchr(subject,' ')-subject]='\0';
+		message=strchr(message,'\n')+1;
+	}
+	if(subject!=NULL)	SockPrintf(socket,"Subject: %s\r\n",subject);
+	SockPrintf(socket,"\r\n%s\r\n",message);
+	footer=readfile(footfile);
+	if(footer!=NULL)	SockPrintf(socket,"%s\r\n",footer);
+	SockPrintf(socket,".\r\n");
+	SockGets(socket,server_reply,4096);
+        puts(server_reply);
+	if(strstr(server_reply,"accepted")!=NULL || strstr(server_reply,"ACCEPTED")!=NULL || strstr(server_reply,"Accepted")!=NULL)
+		success=1;
+	SockPrintf(socket,"QUIT\r\n");
+	SockGets(socket,server_reply,4096);
+	puts(server_reply);
+	SockKill(socket);
+	free(footer);
+	free(server_reply);
+  #endif
   fclose(list);
   free(listfile); free(footfile); free(conffile);
 }
@@ -189,6 +315,7 @@ void subscribe(char *listname, char *user, char *operator)
   FILE *list;
   char *mail,*from;
   char *listfile,*footfile,*welcfile;
+  char *members;
 
   #ifdef LOG
 	if(operator==NULL)
@@ -204,23 +331,37 @@ void subscribe(char *listname, char *user, char *operator)
   footfile=(char *) malloc(strlen(LISTDIR)+strlen(listname)+9);
   sprintf(footfile,"%s/%s.footer",LISTDIR,listname);
 
-  list=fopen(listfile,"a");
-  fprintf(list,"%s\n",user);
-  fclose(list);
-  mail=readfile(welcfile);
-  if(mail!=NULL) {
-	#ifdef LOG
-		fprintf(log,"* sending welcome message\n");
-	#endif
-	if(operator!=NULL)
-		send(listname,user,operator,NULL,mail,footfile);
-	else {
-		from=(char *) malloc(1024);
-		sprintf(from,"%s@%s",listname,HOST);
-		send(listname,user,from,NULL,mail,footfile);
-		free(from);
+  members=readfile(listfile);
+  
+  if(strstr(downcase(members),email(user))!=NULL) { /* User is already subscribed */
+  	mail=(char *) malloc(2048);
+  	if(operator==NULL) { /* User subscribed himheritself */
+  		sprintf(mail,"You are already subscribed to this list.\n");
+	  	send(listname,user,listname,"List subscription...",mail,footfile);	
+  	} else {
+  		sprintf(mail,"You tried to subscibe a user (%s) who is already subscribed to the list.\n",user);
+  		send(listname,operator,listname,"List subscription...",mail,footfile);
+  	}
+  	free(mail);
+  } else {
+	list=fopen(listfile,"a");
+	fprintf(list,"%s\n",user);
+	fclose(list);
+	mail=readfile(welcfile);
+	if(mail!=NULL) {
+		#ifdef LOG
+			fprintf(log,"* sending welcome message\n");
+		#endif
+		if(operator!=NULL)
+			send(listname,user,operator,NULL,mail,footfile);
+		else {
+			from=(char *) malloc(1024);
+			sprintf(from,"%s@%s",listname,HOST);
+			send(listname,user,from,NULL,mail,footfile);
+			free(from);
+		}
+		free(mail);
 	}
-	free(mail);
   }
   free(footfile);  free(welcfile);  free(listfile);
 }
@@ -309,10 +450,24 @@ char *username(char *address)
   char *name;
   name=email(address);
   if ( strchr(name,'@') != NULL )
-    name[strchr(name,'@')-name]='\0';
+  	name[strchr(name,'@')-name]='\0';
   return(name);
 }
 
+char *downcase(char *s)
+{
+  register int i;
+  char *n;
+  
+  n=(char *) malloc(strlen(s)+1);
+  for(i=0;i<=strlen(s);i++)
+  	if(s[i]>='A' && s[i]<='Z')
+  		n[i]=s[i]-'A'+'a';
+  	else
+  		n[i]=s[i];
+  return(n);
+}
+
 int scasecmp(char *s1, char *s2)
 {
   if((s1==NULL || s2==NULL)&&(s1!=NULL || s2!=NULL))
@@ -364,8 +519,8 @@ int main(int argc, char *argv[])
 
   msg=mail+pos+2;
   
-  to=email(extract(header,"To: ",'\n'));
-  to[strchr(to,'@')-to]='\0';
+  to=downcase(email(extract(header,"To: ",'\n')));
+  if(strchr(to,'@')!=NULL) to[strchr(to,'@')-to]='\0';
   
   from=extract(header,"From: ",'\n');
   subject=extract(header,"Subject: ",'\n');
@@ -406,6 +561,20 @@ int main(int argc, char *argv[])
   /* read config file... */
   
   config=readfile(conffile);
+
+  /* Permit "subscribe someone@somewhere" for the user - by converting it
+     to a reasonable form ("subscribe"). */
+     
+  if(strncasecmp(subject,"subscribe ",10)==0) {
+  	recipient=subject+10;
+  	if(scasecmp(email(from),email(recipient))==0)
+  		subject[9]='\0';
+  }
+  if(strncasecmp(subject,"unsubscribe ",12)==0) {
+  	recipient=subject+12;
+  	if(scasecmp(email(from),email(recipient))==0)
+  		subject[11]='\0';
+  }
   
   if(strncasecmp(subject,"subscribe ",10)==0) {
   	recipient=subject+10;
